# -*- coding: utf-8 -*-
#
# Copyright (c) 2018, HILSTER GmbH, Germany - https://hilster.de
# All rights reserved.
#

from __future__ import absolute_import, division, print_function, unicode_literals

from .core import ChecksumMismatchException, _ord
from .typedef import ULInt32


class ChecksumMixin(object):
    """
    Use the static methods encode and decode in your implementation
    and implement _calculate_checksum(data) that returns a calculated checksum.
    """

    def encode(self, full_data=b"", context_data=b""):
        """
        Return the encoded byte type as a binary string.

        Args:
            full_data (bytes): the binary data string until the part to be encoded.
                The user normally does not need to supply this.
            context_data (bytes): the binary data of the current context.
                The user normally does not need to supply this.
        """
        if context_data:
            if self._automatic_calculation:
                checksum = self._calculate_checksum(context_data)
                self.set(checksum)
        return super(self._original_type, self).encode(full_data=full_data, context_data=context_data)

    def decode(self, data, full_data=b"", context_data=b""):
        """
        Decode a binary string into a byte type and return the number
        of bytes that were decoded.

        Args:
            data (bytes): the data buffer that is decoded.
            full_data (bytes): the binary data string until the part to be decoded.
                The user normally does not need to supply this.
            context_data (bytes): the binary data of the current context.
                The user normally does not need to supply this.
        """
        ret = super(self._original_type, self).decode(data, full_data=full_data, context_data=context_data)

        # assert equality if needed
        if self._strict:
            checksum = self._calculate_checksum(context_data)
            if not checksum == super(self._original_type, self).get():
                exceptionFormat = "\\x%%0%ix" % (2 * self._byte_count)
                raise ChecksumMismatchException(
                    value=("Checksum mismatch: expected " +
                           exceptionFormat +
                           " but found " +
                           exceptionFormat) % (checksum, self.get()),
                    object=self)
        return ret

    def set_automatic_calculation(self, enabled=True):
        """
        Enable or disable automatic calculation when ``encode`` is called.

        Args:
            enabled=True (bool): the new enabled state.
        """
        self._automatic_calculation = enabled


class Checksum32(ULInt32, ChecksumMixin):
    """
    Example implementation for ChecksumMixin
    """
    _original_type = ULInt32

    def __init__(self, strict=False, automatic_calculation=True):
        self._strict = strict
        self.set_automatic_calculation(automatic_calculation)
        super(Checksum32, self).__init__()

    def _calculate_checksum(self, data):
        checksum = 0
        for byte in data:
            checksum += _ord(byte)

        return checksum

    def encode(self, full_data=b"", context_data=b""):
        return ChecksumMixin.encode(self, full_data, context_data)

    def decode(self, data, full_data=b"", context_data=b""):
        return ChecksumMixin.decode(self, data, full_data, context_data)

    def _fuzzing_value_iterator(self):
        """
        A generator for all fuzzing values set.
        This sets automatic calculation and strict temporarly to False 
        if fuzzing values are set and restores the original values after the
        fuzzing iterator ends.

        Returns:
            self: if no fuzzing values are set

        Yields:
            fuzzing values
        """
        # store old values
        self._tmp_strict = self._strict
        self._tmp_automatic_calculation = self._automatic_calculation

        if self._fuzzing_values is not None:
            # disable strict and automatic_calculation
            self._strict = False
            self._automatic_calculation = False

        for value in super(self._original_type, self)._fuzzing_value_iterator():
            yield value

        # restore values
        self._strict = self._tmp_strict
        self._automatic_calculation = self._tmp_automatic_calculation
